import gensim
from gensim.models import Word2Vec
from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from gensim.summarization import keywords
from gensim.summarization.summarizer import summarize
from nltk.tokenize import word_tokenize

from aethos.preprocessing.text import process_text


def gensim_textrank_summarizer(
    x_train, x_test=None, list_of_cols=[], new_col_name="_summarized", **algo_kwargs
):
    """
    Uses Gensim Text Rank summarize to extractively summarize text.

    Note this uses a variant of Text Rank.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data that you want to summarize

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `_summarized`
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    for col in list_of_cols:
        x_train.loc[:, col + new_col_name] = [
            summarize(x, **algo_kwargs) for x in x_train[col]
        ]

        if x_test is not None:
            x_test.loc[:, col + new_col_name] = [
                summarize(x, **algo_kwargs) for x in x_test[col]
            ]

    return x_train, x_test


def gensim_textrank_keywords(
    x_train,
    x_test=None,
    list_of_cols=[],
    new_col_name="_extracted_keywords",
    **algo_kwargs
):
    """
    Uses Gensim Text Rank summarize to extract keywords.

    Note this uses a variant of Text Rank.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data that you want to summarize

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `_extracted_keywords`
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    for col in list_of_cols:
        x_train.loc[:, col + new_col_name] = [
            keywords(x, **algo_kwargs) for x in x_train[col]
        ]

        if x_test is not None:
            x_test.loc[:, col + new_col_name] = [
                keywords(x, **algo_kwargs) for x in x_test[col]
            ]

    return x_train, x_test


def gensim_word2vec(x_train, x_test=None, prep=False, col_name=None, **algo_kwargs):
    """
    Uses Gensim Text Rank summarize to extract keywords.

    Note this uses a variant of Text Rank.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    prep : bool, optional
        True to prep the text
        False if text is already prepped.
        By default, False

    col_name : str, optional
        Column name of text data that you want to summarize
        
    Returns
    -------
    Word2Vec
        Word2Vec model
    """

    if prep:
        w2v = Word2Vec(
            sentences=[word_tokenize(process_text(text)) for text in x_train[col_name]],
            **algo_kwargs
        )
    else:
        w2v = Word2Vec(sentences=x_train[col_name], **algo_kwargs)

    return w2v


def gensim_doc2vec(x_train, x_test=None, prep=False, col_name=None, **algo_kwargs):
    """
    Uses Gensim Text Rank summarize to extract keywords.

    Note: this uses a variant of Text Rank.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    prep : bool, optional
        True to prep the text
        False if text is already prepped.
        By default, False

    col_name : str, optional
        Column name of text data that you want to summarize
    
    Returns
    -------
    Doc2Vec
        Doc2Vec Model
    """

    if prep:
        tagged_data = [
            TaggedDocument(words=word_tokenize(process_text(text)), tags=[str(i)])
            for i, text in enumerate(x_train[col_name])
        ]
    else:
        tagged_data = [
            TaggedDocument(words=text, tags=[str(i)])
            for i, text in enumerate(x_train[col_name])
        ]

    d2v = Doc2Vec(tagged_data, **algo_kwargs)
    d2v.delete_temporary_training_data(keep_doctags_vectors=True, keep_inference=True)

    return d2v


def gensim_lda(x_train, x_test=None, prep=False, col_name=None, **algo_kwargs):
    """
    Runs Gensim LDA model and assigns topics to documents.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    prep : bool, optional
        True to prep the text
        False if text is already prepped.
        By default, False

    col_name : str, optional
        Column name of text data that you want to summarize
    
    Returns
    -------
    Dataframe, *Dataframe, LDAModel, corpus, list
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    texts = x_train[col_name].tolist()

    if prep:
        texts = [word_tokenize(process_text(text)) for text in texts]

    id2word = gensim.corpora.Dictionary(texts)
    corpus = [id2word.doc2bow(text) for text in texts]

    lda_model = gensim.models.LdaModel(corpus=corpus, id2word=id2word, **algo_kwargs)

    x_train["topics"] = _assign_topic_doc(lda_model, texts, corpus)

    if x_test is not None:
        texts = x_test[col_name].tolist()

        if prep:
            texts = [word_tokenize(process_text(text)) for text in texts]

        test_corpus = [id2word.doc2bow(text) for text in texts]

        x_test["topics"] = _assign_topic_doc(lda_model, texts, test_corpus)

    return x_train, x_test, lda_model, corpus, id2word


def _assign_topic_doc(lda_model, texts, corpus):
    """
    Helper function to assign the relevant topics to each document
    
    Parameters
    ----------
    lda_model : LDAModel
        LDA Model

    texts : [str]
        List of text documents

    corpus : list
        Corpus list
    
    Returns
    -------
    list
        List of topics assigned to each document
    """

    keywords = []

    for i, row in enumerate(lda_model[corpus]):

        row = sorted(row, key=lambda x: (x[1]), reverse=True)

        for j, (topic_num, prop_topic) in enumerate(row):
            if j == 0:
                wp = lda_model.show_topic(topic_num)
                topic_keywords = ", ".join([word for word, prop in wp])
                keywords.append(topic_keywords)

                break

    return keywords



import collections
import os
from collections import OrderedDict

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder

DATA_CHECKLIST = {
    "Convert files to .csv",
    "Merge files",
    "Fix encoding issues",
    "Clean column names (english, no whitespace, no special chars)",
    "Are there duplicate columns?",
    "Fix datatypes (datetime, int, float, string)",
}

CLEANING_CHECKLIST = {
    "Non-sensical observations/artifacts?",
    "Coding of categorical features?",
    "Missing values?",
    "Outliers?",
    "Constant values (=Zero Importance)?",
    "Low importance features?",
    "Collinear, correlated or otherwise dependent features?",
    "Highly skewed features?",
    "Irrelevant features?",
}

UNI_ANALYSIS_CHECKLIST = {
    "Look at mean, median, min, max, std, iqr, quantiles (1%, 5%, 25%, 50%, 75%, 95%, 99%)",
    "Draw boxplots, histograms",
}

MULTI_ANALYSIS_CHECKLIST = {"Draw scatter plots", "Create correlation matrix"}

ISSUES_CHECKLIST = {
    "Impute missing values (mode, median, mean)",
    "Remove variables that have too many missings",
    "Remove observations that have too many missings",
    "Select appropriate time slice",
}

PREPARATION_CHECKLIST = {
    "Clip values that are too small/too large",
    "Scale to [0,1] or normalize (mean=0, std=1) or Robust / Quantile Scaling",
    "One-hot encoding, Label Encoding (0,1,2,3)",
    "Create log-transformed versions for highly skewed variables",
    "Create binned versions for variables",
    "Combine categories for highly skewed categorical variables",
    "Create sum/difference/product/quotient of variables",
    "Create polynomial features",
}


def label_encoder(x_train, x_test=None, list_of_cols=[], target=False):
    """
    Label encodes the columns provided.
    
    Either the full data or training data plus testing data MUST be provided, not both.
    
    Parameters
    ----------
    x_train: Dataframe or array like - 2d
        Dataset.

    x_test: Dataframe or array like - 2d
        Testing dataset, by default None.

    list_of_cols : list, optional
        A list of specific columns to apply this technique to
        If `list_of_cols` is not provided, the strategy will be
        applied to all numeric columns., by default []

    target : bool, optional
        True if the column being encoded is the target variable, by default False

    Returns
    -------
    Dataframe, *Dataframe, dict
        Transformed dataframe with rows with a missing values in a specific column are missing

    Returns 2 Dataframes x_test is provided.  
    """

    label_encode = LabelEncoder()
    target_mapping = None

    for col in list_of_cols:
        x_train[col] = label_encode.fit_transform(x_train[col])

        if x_test is not None:
            x_test[col] = label_encode.transform(x_test[col])

    if target:
        target_mapping = dict(
            zip(range(len(label_encode.classes_)), label_encode.classes_)
        )

    return x_train, x_test, target_mapping


def check_missing_data(df) -> bool:
    """
    Utility function that checks if the data has any missing values.
    
    Parameters
    ----------
    df : Dataframe
        Pandas Dataframe of data
    
    Returns
    -------
    bool
        True if data has missing values, False otherwise.
    """

    return df.isnull().values.any()


def get_keys_by_values(dict_of_elements: dict, item) -> list:
    """
    Utility function that returns the list of keys whos value matches a criteria defined
    by the param `item`.
    
    Parameters
    ----------
    dict_of_elements : dict
        Dictionary of key value mapping

    item : Any
        Value you want to return keys of.
    
    Returns
    -------
    list
        List of keys whos value matches the criteria defined by the param `item`
    """

    return [key for (key, value) in dict_of_elements.items() if value == item]


def drop_replace_columns(df, drop_cols, new_data, keep_col=False):
    """
    Utility function that drops a column that has been processed and replaces it with the new columns that have been derived from it.
    
    Parameters
    ----------
    df : Dataframe
        Dataframe of the data

    drop_cols : str or [str]
        Column or list of columns to be dropped

    new_data : Dataframe
        New data columns to be added to the dataframe
    
    Returns
    -------
    Dataframe
        Dataframe with the dropped column and the new data added
    """

    if keep_col:
        df = pd.concat([df, new_data], axis=1)
    else:
        df = df.drop(drop_cols, axis=1)
        df = pd.concat([df, new_data], axis=1)

    return df


def split_data(df, split_percentage: float):
    """
    Function that splits the data into a training and testing set. Split percentage is passed in through
    the split_percentage variable.
    
    Parameters
    ----------
    df : Dataframe
        Full dataset you want to split.

    split_percentage : float
        The % of data that you want in your test set, split_percentage is the percentage of data in the traning set.
    
    Returns
    -------
    Dataframe, Dataframe
        Train data and test data.
    """

    x_train, x_test = train_test_split(df, test_size=split_percentage)

    return x_train, x_test


def _numeric_input_conditions(list_of_cols: list, x_train) -> list:
    """
    Helper function to help set variable values of numeric cleaning method functions.

    If list of columns is provided, use it.

    If list of columns is not provided, use all the columns.
    """

    if list_of_cols:
        list_of_cols = list_of_cols
    else:
        list_of_cols = x_train.select_dtypes([np.number]).columns.tolist()

    return list_of_cols


def _get_columns(list_of_cols, x_train) -> list:
    """
    If the list of columns are optional and no columns are provided, the columns are set
    to all the columns in the data.
    
    Parameters
    ----------
    list_of_cols : list
        List of columns to apply method to.

    x_train : Dataframe or array like - 2d
        Training Dataset
    
    Returns
    -------
    list
        List of columns to apply technique to
    """

    if list_of_cols:
        list_of_cols = list_of_cols
    else:
        list_of_cols = x_train.columns.tolist()

    return list_of_cols


def _input_columns(list_args: list, list_of_cols: list):
    """
    Takes columns inputted as arguments vs. columns passed as a list
    and returns a final column list.
    
    Parameters
    ----------
    list_args : list
        Input columns passed as args

    list_of_cols : list
        Hardcoded provided list of input columns.
    
    Returns
    -------
    list
        List of columns in a dataset
    """

    if list_of_cols or (not list_of_cols and not list_args):
        column_list = list_of_cols
    else:
        column_list = list(list_args)

    return column_list


def _set_item(
    x_train, x_test, column: str, value: list, train_length: int, test_length: int
):
    """
    Utility function for __setitem__ for determining which input is for which dataset
    and then sets the input to the new column for the correct dataset.
    
    Parameters
    ----------
    x_train : Dataframe
        Training Data

    x_test : Dataframe
        Testing Data

    column : str
        New column name

    value : list
        List of values for new column

    train_length : int
        Length of training data
        
    test_length : int
        Length of training data
    """

    ## If the training data and testing data have the same number of rows, apply the value to both
    ## train and test data set
    if len(value) == train_length and len(value) == test_length:
        x_train[column] = value
        x_test[column] = value
    elif len(value) == train_length:
        x_train[column] = value
    else:
        x_test[column] = value

    return x_train, x_test


def _make_dir(path: str):
    """
    Makes directory if it does exist.
    
    Parameters
    ----------
    path : str
        Path of directory or file
    """

    if not os.path.exists(path):
        os.makedirs(path)



from aethos.config import cfg, DEFAULT_IMAGE_DIR
from aethos.util import _make_dir


def _make_image_dir():  # pragma: no cover

    if not cfg["images"]["dir"]:
        image_dir = DEFAULT_IMAGE_DIR
    else:
        image_dir = cfg["images"]["dir"]

    _make_dir(image_dir)

    return image_dir



import unittest
import shutil
from pathlib import Path

import numpy as np
import pandas as pd

from aethos import Data


class TestPreprocessing(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(str(Path.home()) + "/.aethos/reports/")

    def test_preprocessnumeric_normalize(self):

        unnormal_data = [[5.0, 3, 1], [2.0, 2, 1], [10.0, 1, 1], [10.0, 1, 1]]

        columns = ["col1", "col2", "col3"]
        data = pd.DataFrame(unnormal_data, columns=columns)

        preprocess = Data(x_train=data, test_split_percentage=0.5, report_name="test")
        preprocess.normalize_numeric(keep_col=False)
        validate = preprocess.x_train.values.tolist()

        self.assertTrue(True)

    def test_preprocessnumeric_robust(self):

        unnormal_data = [[5.0, 3, 1], [2.0, 2, 1], [10.0, 1, 1], [10.0, 1, 1]]

        columns = ["col1", "col2", "col3"]
        data = pd.DataFrame(unnormal_data, columns=columns)

        preprocess = Data(x_train=data, test_split_percentage=0.5, report_name="test")
        preprocess.normalize_quantile_range(keep_col=False)
        validate = preprocess.x_train.values.tolist()

        self.assertTrue(True)

    def test_preprocessnumeric_log(self):

        unnormal_data = [[1.0, -2.0, 2.0], [-2.0, 1.0, 3.0], [4.0, 1.0, -2.0]]

        columns = ["col1", "col2", "col3"]
        data = pd.DataFrame(unnormal_data, columns=columns)

        preprocess = Data(x_train=data, test_split_percentage=0.5, report_name="test")
        preprocess.normalize_log()
        preprocess.normalize_log(base=2)
        preprocess.normalize_log(base=10)

        self.assertTrue(True)

    def test_preprocess_traindata(self):

        unnormal_x_train = [[5.0, 3, 1], [2.0, 2, 1], [10.0, 1, 1]]

        unnormal_x_test = [[5.0, 3, 1], [2.0, 2, 1], [10.0, 1, 1]]

        columns = ["col1", "col2", "col3"]
        x_train = pd.DataFrame(unnormal_x_train, columns=columns)
        x_test = pd.DataFrame(unnormal_x_test, columns=columns)

        preprocess = Data(
            x_train=x_train,
            x_test=x_test,
            test_split_percentage=0.5,
            report_name="test",
        )
        preprocess.normalize_numeric("col1", "col2", "col3")
        validate_train = preprocess.x_train.values.tolist()
        validate_test = preprocess.x_test.values.tolist()

        self.assertListEqual(validate_train, validate_test)

    def test_preprocess_splitsentences(self):

        text_data = [
            "Hi my name is aethos. Please split me.",
            "This function is going to split by sentence. Automation is great.",
        ]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.split_sentences("data")
        validate = prep.x_train["data_sentences"].values.tolist()

        self.assertListEqual(
            validate,
            [
                ["Hi my name is aethos.", "Please split me."],
                [
                    "This function is going to split by sentence.",
                    "Automation is great.",
                ],
            ],
        )

    def test_preprocess_nltkstem(self):

        text_data = [
            "Hi my name is aethos. Please split me.",
            "This function is going to split by sentence. Automation is great.",
        ]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.stem_nltk("data")
        validate = prep.x_train.shape[1]

        self.assertEqual(validate, 2)

    def test_preprocess_nltksplit(self):

        text_data = ["Please.exe split me."]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.split_words_nltk("data")
        validate = prep.x_train.data_tokenized.values.tolist()

        self.assertListEqual(validate, [["Please.exe", "split", "me", "."]])

    def test_preprocess_nltksplit_regex(self):

        text_data = ["Please123 split me."]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.split_words_nltk("data", regexp=r"\w+\d+")
        validate = prep.x_train.data_tokenized.values.tolist()

        self.assertListEqual(validate, [["Please123"]])

    def test_preprocess_nltkremove_punctuation(self):

        text_data = ["Please split me."]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.remove_punctuation("data")
        validate = prep.x_train.data_rem_punct.values.tolist()

        self.assertListEqual(validate, ["Please split me"])

    def test_preprocess_nltkremove_punctuation_regexp(self):

        text_data = ["Please.exe, split me.", "hello it's me, testing.dll."]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.remove_punctuation("data", regexp=r"\w+\.\w+|\w+")
        validate = prep.x_train.data_rem_punct.values.tolist()

        self.assertListEqual(
            validate, ["Please.exe split me", "hello it s me testing.dll"]
        )

    def test_preprocess_cleantext(self):

        text_data = ["Please.exe, split me.", "hello it's me, testing.dll."]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.clean_text("data")
        validate = prep.x_train["data_clean"].tolist()

        self.assertListEqual(validate, ["please.ex split", "hello 's testing.dl"])

    def test_preprocess_nltkremove_punctuation_exception(self):

        text_data = ["Please,> split me."]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.remove_punctuation("data", exceptions=[".", ">"])
        validate = prep.x_train.data_rem_punct.values.tolist()

        self.assertListEqual(validate, ["Please> split me."])

    def test_preprocess_nltkremove_stopwords(self):

        text_data = ["Please the split me."]
        data = pd.DataFrame(data=text_data, columns=["data"])

        prep = Data(x_train=data, split=False, report_name="test")
        prep.remove_stopwords_nltk("data", custom_stopwords=["please"])
        validate = prep.x_train.data_rem_stop.values.tolist()

        self.assertListEqual(validate, ["split ."])


if __name__ == "__main__":
    unittest.main()



"""
This file contains the following methods:

feature_bag_of_words
feature_hashing_vectorizer
feature_tfidf
nltk_feature_postag
"""

import pandas as pd
import spacy
from sklearn.feature_extraction.text import (
    CountVectorizer,
    HashingVectorizer,
    TfidfVectorizer,
)
from textblob import TextBlob

from aethos.util import _get_columns, drop_replace_columns


def feature_bag_of_words(
    x_train, x_test=None, list_of_cols=[], keep_col=False, **algo_kwargs
):
    """
    Creates a matrix of how many times a word appears in a document.
    
    Parameters
    ----------
    x_train : DataFrame
        Training dataset, by default None
        
    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to., by default []

    keep_col : bool, optional
        True if you want to keep the columns passed, otherwise remove it.

    algo_kwargs : dict, optional
        Parameters you would pass into Bag of Words constructor as a dictionary., by default {}
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    enc = CountVectorizer(**algo_kwargs)
    list_of_cols = _get_columns(list_of_cols, x_train)

    for col in list_of_cols:
        enc_data = enc.fit_transform(x_train[col]).toarray()
        enc_df = pd.DataFrame(enc_data, columns=enc.get_feature_names())
        x_train = drop_replace_columns(x_train, col, enc_df, keep_col)

        if x_test is not None:
            enc_x_test = enc.transform(x_test[col]).toarray()
            enc_test_df = pd.DataFrame(enc_x_test, columns=enc.get_features_names())
            x_test = drop_replace_columns(x_test, col, enc_test_df, keep_col)

    return x_train, x_test


def feature_hash_vectorizer(
    x_train, x_test=None, list_of_cols=[], keep_col=True, **hashing_kwargs
):
    """
    Returns a hashed encoding of text data.
    
    Parameters
    ----------
    x_train : DataFrame
        Training dataset, by default None
        
    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to., by default []

    keep_col : bool, optional
        True if you want to keep the columns passed, otherwise remove it.

    hashing_kwargs : dict, optional
        Parameters you would pass into Hashing Vectorizer constructor, by default {}
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    enc = HashingVectorizer(**hashing_kwargs)
    list_of_cols = _get_columns(list_of_cols, x_train)

    for col in list_of_cols:
        enc_data = enc.fit_transform(x_train[col]).toarray()
        enc_df = pd.DataFrame(enc_data)
        x_train = drop_replace_columns(x_train, col, enc_df, keep_col)

        if x_test is not None:
            enc_x_test = enc.transform(x_test[col]).toarray()
            enc_test_df = pd.DataFrame(enc_x_test)
            x_test = drop_replace_columns(x_test, col, enc_test_df, keep_col)

    return x_train, x_test


def feature_tfidf(x_train, x_test=None, list_of_cols=[], keep_col=True, **algo_kwargs):
    """
    Creates a matrix of the tf-idf score for every word in the corpus as it pertains to each document.
    
    Either the full data or training data plus testing data MUST be provided, not both.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset
        
    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to, by default []

    keep_col : bool, optional
        True if you want to keep the columns passed, otherwise remove it.

    algo_kwargs :  optional
        Parameters you would pass into TFIDF constructor, by default {}
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column

    Returns 2 Dataframes if x_test data is provided. 
    """

    enc = TfidfVectorizer(**algo_kwargs)
    list_of_cols = _get_columns(list_of_cols, x_train)

    for col in list_of_cols:
        enc_data = enc.fit_transform(x_train[col]).toarray()
        enc_df = pd.DataFrame(enc_data, columns=enc.get_feature_names())
        x_train = drop_replace_columns(x_train, col, enc_df, keep_col)

        if x_test is not None:
            enc_x_test = enc.transform(x_test[col]).toarray()
            enc_test_df = pd.DataFrame(enc_x_test, columns=enc.get_feature_names())
            x_test = drop_replace_columns(x_test, col, enc_test_df, keep_col)

    return x_train, x_test


def nltk_feature_postag(
    x_train, x_test=None, list_of_cols=[], new_col_name="_postagged"
):
    """
    Part of Speech tag the text data provided. Used to tag each word as a Noun, Adjective,
    Verbs, etc.

    This utilizes TextBlob which utlizes the NLTK tagger and is a wrapper for the tagging process.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to, by default []

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_postagged`
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    list_of_cols = _get_columns(list_of_cols, x_train)

    for col in list_of_cols:
        x_train[col + new_col_name] = pd.Series([TextBlob(x) for x in x_train[col]])

        if x_test is not None:
            x_test[col + new_col_name] = pd.Series(
                [TextBlob(x).tags for x in x_test[col]]
            )

    return x_train, x_test


def nltk_feature_noun_phrases(
    x_train, x_test=None, list_of_cols=[], new_col_name="_phrases"
):
    """
    Extracts noun phrases from the given text.

    This utilizes TextBlob which utlizes the NLTK NLP engine.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to, by default []

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_phrases`
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    list_of_cols = _get_columns(list_of_cols, x_train)

    for col in list_of_cols:
        x_train[col + new_col_name] = pd.Series(
            [TextBlob(x).noun_phrases for x in x_train[col]]
        )

        if x_test is not None:
            x_test[col + new_col_name] = pd.Series(
                [TextBlob(x).noun_phrases for x in x_train[col]]
            )

    return x_train, x_test


# TODO: Add simple and complex spacy pos tagging
def spacy_feature_postag(
    x_train, x_test=None, list_of_cols=[], new_col_name="_postagged"
):
    """
    Part of Speech tag the text data provided. Used to tag each word as a Noun, Adjective,
    Verbs, etc.

    This utilizes the spacy NLP engine.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to, by default []

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_postagged`
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    list_of_cols = _get_columns(list_of_cols, x_train)

    nlp = spacy.load("en_core_web_sm")

    for col in list_of_cols:
        transformed_text = map(nlp, x_train[col])
        x_train[col + new_col_name] = pd.Series(
            map(lambda x: [(token, token.pos_) for token in x], transformed_text,)
        )

        if x_test is not None:
            transformed_text = map(nlp, x_test[col])
            x_test[col + new_col_name] = pd.Series(
                map(lambda x: [(token, token.pos_) for token in x], transformed_text,)
            )

    return x_train, x_test


# TODO: Double check spacy noun phrase implementation
def spacy_feature_noun_phrases(
    x_train, x_test=None, list_of_cols=[], new_col_name="_phrases"
):
    """
    Extracts noun phrases from the given data.

    This utilizes the spacy NLP engine.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to, by default []

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_phrases`
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    list_of_cols = _get_columns(list_of_cols, x_train)

    nlp = spacy.load("en")

    for col in list_of_cols:
        transformed_text = list(map(nlp, x_train[col]))
        x_train[col + new_col_name] = pd.Series(
            map(lambda x: [str(phrase) for phrase in x.noun_chunks], transformed_text)
        )

        if x_test is not None:
            transformed_text = map(nlp, x_test[col])
            x_test[col + new_col_name] = pd.Series(
                map(
                    lambda x: [str(phrase) for phrase in x.noun_chunks],
                    transformed_text,
                )
            )

    return x_train, x_test



"""
This file contains the following methods:

replace_missing_mean_median_mode
replace_missing_constant
"""

import pandas as pd
from aethos.cleaning.categorical import replace_missing_new_category
from aethos.util import _get_columns, _numeric_input_conditions, drop_replace_columns
from sklearn.impute import SimpleImputer
import numpy as np


def replace_missing_mean_median_mode(
    x_train, x_test=None, list_of_cols=[], strategy=""
):
    """
    Replaces missing values in every numeric column with the mean, median or mode of that column specified by strategy.

    Mean: Average value of the column. Effected by outliers.
    Median: Middle value of a list of numbers. Equal to the mean if x_train follows normal distribution. Not effected much by anomalies.
    Mode: Most common number in a list of numbers.
    
    Parameters
    ----------
    x_train: Dataframe or array like - 2d
        Dataset

    x_test: Dataframe or array like - 2d
        Testing dataset, by default None.

    list_of_cols : list, optional
        A list of specific columns to apply this technique to
        If `list_of_cols` is not provided, the strategy will be
        applied to all numeric columns., by default []

    strategy : str
        Strategy for replacing missing values.
        Can be either "mean", "median" or "most_frequent"
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with rows with a missing values in a specific column are missing

    Returns 2 Dataframes test if x_test is provided.  
    """

    if strategy != "most_frequent":
        list_of_cols = _numeric_input_conditions(list_of_cols, x_train)
    else:
        list_of_cols = _get_columns(list_of_cols, x_train)

    imp = SimpleImputer(strategy=strategy)

    fit_data = imp.fit_transform(x_train[list_of_cols])
    fit_df = pd.DataFrame(fit_data, columns=list_of_cols)
    x_train = drop_replace_columns(x_train, list_of_cols, fit_df)

    if x_test is not None:
        fit_x_test = imp.transform(x_test[list_of_cols])
        fit_test_df = pd.DataFrame(fit_x_test, columns=list_of_cols)
        x_test = drop_replace_columns(x_test, list_of_cols, fit_test_df)

    return x_train, x_test


def replace_missing_constant(x_train, x_test=None, col_to_constant=None, constant=0):
    """
    Replaces missing values in every numeric column with a constant. If `col_to_constant` is not provided,
    all the missing values in the x_train will be replaced with `constant`
    
    Parameters
    ----------
    col_to_constant : list, dict, optional
        Either a list of columns to replace missing values or a `column`: `value` dictionary mapping,
        by default None

    constant : int, float, optional
        Value to replace missing values with, by default 0

    x_train: Dataframe or array like - 2d
        Training dataset, by default None.
        
    x_test: Dataframe or array like - 2d
        Testing dataset, by default None.
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with rows with a missing values in a specific column are missing

    Returns 2 Dataframes if x_test is provided.  
    
    Examples
    ------
    >>> replace_missing_constant({'a': 1, 'b': 2, 'c': 3})
    >>> replace_missing_constant(1, ['a', 'b', 'c'])
    """

    if isinstance(col_to_constant, dict):
        x_train, x_test = replace_missing_new_category(
            col_to_cateogory=col_to_constant, x_train=x_train, x_test=x_test
        )
    elif isinstance(col_to_constant, list):
        x_train, x_test = replace_missing_new_category(
            constant=constant,
            col_to_category=col_to_constant,
            x_train=x_train,
            x_test=x_test,
        )
    else:
        x_train, x_test = replace_missing_new_category(
            constant=constant, x_train=x_train, x_test=x_test
        )

    return x_train, x_test
